__license__ = "GNU General Public License v3"
__author__ = 'Hwaipy'
__email__ = 'hwaipy@gmail.com'

import threading
import msgpack
import re
from tornado.ioloop import IOLoop
from threading import Thread
import types


class IFDefinition:
    PROTOCOL = b'IF1'
    DISTRIBUTING_MODE_BROKER = b'Broker'
    DISTRIBUTING_MODE_DIRECT = b'Direct'
    DISTRIBUTING_MODE_SERVICE = b'Service'
    HEARTBEAT_LIVETIME = 10


class IFException(Exception):
    def __init__(self, description):
        Exception.__init__(self)
        self.description = description

    def __str__(self):
        return self.description


class IFLoop:
    __loopLock = threading.Lock()
    __running = False
    __loopingThread = None

    @classmethod
    def start(cls, background=True):
        if not IFLoop.__runIfNot(): raise IFException('IFLoop is already running.')
        if background:
            thread = Thread(target=IOLoop.current().start)
            thread.setDaemon(True)
            thread.start()
            IFLoop.__loopingThread = thread
        else:
            IFLoop.__loopingThread = threading.current_thread()
            IOLoop.current().start()

    @classmethod
    def tryStart(cls):
        if IFLoop.__runIfNot():
            thread = Thread(target=IOLoop.current().start)
            thread.setDaemon(True)
            thread.start()
            IFLoop.__loopingThread = thread

    @classmethod
    def __runIfNot(cls):
        IFLoop.__loopLock.acquire()
        if IFLoop.__running:
            IFLoop.__loopLock.release()
            return False
        IFLoop.__running = True
        IFLoop.__loopLock.release()
        return True

    @classmethod
    def join(cls):
        IFLoop.__loopingThread.join()


class Message:
    MessageIDs = 0

    @classmethod
    def newBrokerMessage(cls, invocation, serialization='Msgpack'):
        return Message.newMessage(IFDefinition.DISTRIBUTING_MODE_BROKER, b'', invocation, serialization)

    @classmethod
    def newServiceMessage(cls, serviceName, invocation, serialization='Msgpack'):
        return Message.newMessage(IFDefinition.DISTRIBUTING_MODE_SERVICE, serviceName, invocation, serialization)

    @classmethod
    def newDirectMessage(cls, address, invocation, serialization='Msgpack'):
        return Message.newMessage(IFDefinition.DISTRIBUTING_MODE_DIRECT, address, invocation, serialization)

    @classmethod
    def newMessage(cls, distributingMode, distributingAddress, invocation, serialization='Msgpack'):
        if (distributingMode != IFDefinition.DISTRIBUTING_MODE_BROKER) and (
                distributingMode != IFDefinition.DISTRIBUTING_MODE_DIRECT) and (
                distributingMode != IFDefinition.DISTRIBUTING_MODE_SERVICE): raise IFException(
            'Bad DistributingMode: {}'.format(distributingMode))
        if distributingMode == IFDefinition.DISTRIBUTING_MODE_BROKER: distributingAddress = b''
        # if serialization == 'Plain' or serialization == 'Default' or serialization == 'ZMQ': serialization = b''
        id = Message.__getAndIncrementID()
        id = str(id)
        msg = [b'', IFDefinition.PROTOCOL, id, distributingMode, distributingAddress,
               serialization, invocation.serialize(serialization)]
        return Message([Message.__messagePartToBytes(m) for m in msg])

    @classmethod
    def newFromBrokerMessage(cls, fromAddress, invocation, serialization='Msgpack'):
        id = Message.__getAndIncrementID()
        id = str(id)
        msg = [b'', IFDefinition.PROTOCOL, id, fromAddress, serialization, invocation.serialize(serialization)]
        return [Message.__messagePartToBytes(m) for m in msg]

    @classmethod
    def __getAndIncrementID(cls):
        __mutex__ = threading.Lock()
        __mutex__.acquire()
        id = Message.MessageIDs
        Message.MessageIDs += 1
        __mutex__.release()
        return id

    @classmethod
    def __messagePartToBytes(self, p):
        if isinstance(p, bytes): return p
        if isinstance(p, str): return bytes(p, 'UTF-8')
        if p == None: return b''
        raise IFException('Data type not transportable: {}'.format(type(p)))

    def __init__(self, msgc):
        self.__content = msgc
        self.__protocol = msgc[1]
        self.messageID = msgc[2].decode('UTF-8')
        if self.isOutgoingMessage():
            self.__distributingMode = msgc[3]
            self.distributingAddress = msgc[4]
            self.serialization = msgc[5]
            self.__invocationContent = msgc[6]
        else:
            self.__distributingMode = b'Received'
            self.fromAddress = msgc[3]
            self.serialization = msgc[4]
            self.__invocationContent = msgc[5]
        self.__invocation = None

    def isProtocolValid(self):
        return self.__protocol == IFDefinition.PROTOCOL

    def isOutgoingMessage(self):
        return len(self.__content) == 7

    def isBrokerMessage(self):
        return self.isOutgoingMessage() and self.__distributingMode == IFDefinition.DISTRIBUTING_MODE_BROKER

    def isServiceMessage(self):
        return self.isOutgoingMessage() and self.__distributingMode == IFDefinition.DISTRIBUTING_MODE_SERVICE

    def isDirectMessage(self):
        return self.isOutgoingMessage() and self.__distributingMode == IFDefinition.DISTRIBUTING_MODE_DIRECT

    def getInvocation(self, decoded=True):
        if not decoded:
            return self.__invocationContent
        if not self.__invocation:
            self.__invocation = Invocation.deserialize(self.__invocationContent, self.serialization)
        return self.__invocation

    def getContent(self):
        return self.__content

    def __str__(self):
        if self.isBrokerMessage():
            return 'Broker: [id={}] {}'.format(self.messageID, self.getInvocation())
        if self.isServiceMessage():
            return 'Service [{}]: [id={}] {}'.format(self.distributingAddress, self.messageID, self.getInvocation())
        if self.isDirectMessage():
            return 'Direct [{}]: [id={}] {}'.format(self.distributingAddress, self.messageID, self.getInvocation())
        else:
            return 'Receive from [{}]: [id={}] {}'.format('Broker' if self.fromAddress == b'' else self.fromAddress,
                                                          self.messageID, self.getInvocation())


class Invocation:
    KeyType = u'Type'
    KeyFunciton = u'Function'
    KeyArguments = u'Arguments'
    KeyKeyworkArguments = u'KeyworkArguments'
    KeyRespopnseID = u'ResponseID'
    KeyResult = u'Result'
    KeyError = u'Error'
    KeyWarning = u'Warning'
    ValueTypeRequest = u'Request'
    ValueTypeResponse = u'Response'
    Preserved = [KeyType, KeyFunciton, KeyArguments, KeyKeyworkArguments, KeyRespopnseID, KeyResult, KeyError,
                 KeyWarning]

    def __init__(self, content={}):
        self.__content = content

    def get(self, key, nilValid=True, nonKeyValid=True):
        if self.__content.__contains__(key):
            value = self.__content[key]
            if value == None:
                if nilValid:
                    return None
                else:
                    raise IFException("Nil value invalid with key {}.".format(key))
            else:
                return value
        elif (nonKeyValid):
            return None
        else:
            raise IFException("Invocation does not contains key {}.".format(key))

    def isRequest(self):
        return self.get(Invocation.KeyType) == Invocation.ValueTypeRequest

    def isResponse(self):
        return self.get(Invocation.KeyType) == Invocation.ValueTypeResponse

    def isError(self):
        return self.isResponse() and self.__content.__contains__(Invocation.KeyError)

    def hasWarning(self):
        return self.isResponse() and self.__content.__contains__(Invocation.KeyWarning)

    def getResponseID(self):
        if self.isResponse(): return self.get(Invocation.KeyRespopnseID)
        return None

    def getResult(self):
        if self.isResponse() and not self.isError():
            return self.get(Invocation.KeyResult)
        return None

    def getError(self):
        if self.isError(): return self.get(Invocation.KeyError)
        return None

    def getWarning(self):
        if self.isResponse(): return self.get(Invocation.KeyWarning)
        return None

    def getFunction(self):
        if self.isRequest():
            return self.get(Invocation.KeyFunciton)
        return None

    def getArguments(self):
        if self.isRequest():
            args = self.get(Invocation.KeyArguments)
            if args: return args if isinstance(args, list) else [args]
            return []
        return None

    def getKeywordArguments(self):
        if self.isRequest():
            kwargs = self.get(Invocation.KeyKeyworkArguments)
            return kwargs if kwargs else {}
        return None

    @classmethod
    def newRequest(cls, functionName, args, kwargs):
        return Invocation({
            Invocation.KeyType: Invocation.ValueTypeRequest,
            Invocation.KeyFunciton: functionName,
            Invocation.KeyArguments: args,
            Invocation.KeyKeyworkArguments: kwargs
        })

    @classmethod
    def newResponse(cls, messageID, result):
        return Invocation({
            Invocation.KeyType: Invocation.ValueTypeResponse,
            Invocation.KeyRespopnseID: messageID,
            Invocation.KeyResult: result
        })

    @classmethod
    def newError(cls, messageID, description):
        return Invocation({
            Invocation.KeyType: Invocation.ValueTypeResponse,
            Invocation.KeyRespopnseID: messageID,
            Invocation.KeyError: description
        })

    def serialize(self, serialization='Msgpack'):
        if serialization == 'Msgpack':
            return msgpack.packb(self.__content, use_bin_type=True)
        else:
            raise IFException('Bad serialization: {}'.format(serialization))

    def __str__(self):
        content = ', '.join(['{}: {}'.format(k, self.__content[k]) for k in self.__content.keys()])
        return "Invocation [{}]".format(content)

    @classmethod
    def deserialize(cls, data, serialization='Msgpack', contentOnly=False):
        if isinstance(serialization, bytes): serialization = str(serialization, encoding='UTF-8')
        if serialization == 'Msgpack':
            unpacker = msgpack.Unpacker(raw=False)
            unpacker.feed(data)
            content = unpacker.__next__()
            return content if contentOnly else Invocation(content)
        else:
            raise IFException('Bad serialization: {}'.format(serialization))

    # sourcePoint only available for Broker.
    async def perform(self, target, sourcePoint=None):
        try:
            method = getattr(target, self.getFunction())
        except BaseException as e:
            raise IFException('Function [{}] not available.'.format(self.getFunction()))
        if not callable(method):
            raise IFException('Function [{}] not available.'.format(self.getFunction()))
        args = self.getArguments()
        kwargs = self.getKeywordArguments()
        if sourcePoint:
            args = [sourcePoint] + args
        try:
            result = method(*args, **kwargs)
            if (isinstance(result, types.CoroutineType)):
                return await result
            else:
                return result
        except BaseException as e:
            matchEargs = re.search("takes ([0-9]+) positional arguments but ([0-9]+) were given", str(e))
            if matchEargs:
                minus = 2 if sourcePoint else 1
                raise IFException('Function [{}] expects [{}] arguments, but [{}] were given.'
                                  .format(self.getFunction(),
                                          int(matchEargs.group(1)) - minus,
                                          int(matchEargs.group(2)) - minus))
            matchEkwargs = re.search("got an unexpected keyword argument '(.+)'", str(e))
            if matchEkwargs:
                raise IFException(
                    'Keyword Argument [{}] not availabel for function [{}].'.format(matchEkwargs.group(1),
                                                                                    self.getFunction()))
            raise e

        # noResponse = message.get(Message.KeyNoResponse)
        # if callable(method):
        #     try:
        #         result = method(*args, **kwargs)
        #         response = message.response(result)
        #         if noResponse is not True:
        #             self.communicator.sendLater(response)
        #     except BaseException as e:
        #         error = message.error(e.__str__())
        #         self.communicator.sendLater(error)
        #     return




import zmq
from zmq.eventloop.zmqstream import ZMQStream
from tornado.ioloop import IOLoop
from interactionfreepy.core import IFException, Message, Invocation, IFLoop, IFDefinition
import time
import threading
import asyncio
import logging

class IFWorker(object):
    def __init__(self, endpoint, serviceName=None, serviceObject=None, interfaces=[], blocking=True, timeout=None):
        self.__endpoint = endpoint
        socket = zmq.Context().socket(zmq.DEALER)
        self.__stream = ZMQStream(socket, IOLoop.current())
        self.__stream.on_recv(self.__onMessage)
        self.__stream.socket.setsockopt(zmq.LINGER, 0)
        self.__stream.connect(self.__endpoint)
        self.__waitingMap = {}
        self.__waitingMapLock = threading.Lock()
        self.blocking = blocking
        self.timeout = timeout
        self.__isService = False
        if serviceName != None:
            self.bindService(serviceName, serviceObject, interfaces)
        threading.Thread(target=self.__hbLoop, daemon=True).start()
        IFLoop.tryStart()
        if self.__isService:
            self.registerAsService(self.__serviceName, self.__interfaces)

    def __hbLoop(self):
        while True:
            time.sleep(IFDefinition.HEARTBEAT_LIVETIME / 5)
            try:
                if not self.blockingInvoker(timeout=IFDefinition.HEARTBEAT_LIVETIME / 5).heartbeat() \
                        and self.__isService:
                    self.registerAsService(self.__serviceName, self.__interfaces)
            except BaseException as e:
                logging.warning('Heartbeat: {}'.format(e))

    def __onMessage(self, msg):
        try:
            message = Message(msg)
            invocation = message.getInvocation()
            if invocation.isRequest():
                IOLoop.current().add_callback(self.__onRequest, message)
            elif invocation.isResponse():
                self.__onResponse(message)
        except BaseException as e:
            import traceback
            exstr = traceback.format_exc()
            logging.debug(exstr)

    async def __onRequest(self, message):
        sourcePoint = message.fromAddress
        invocation = message.getInvocation()
        try:
            result = await invocation.perform(self.__serviceObject)
            responseMessage = Message.newDirectMessage(sourcePoint, Invocation.newResponse(message.messageID, result))
            self.__stream.send_multipart(responseMessage.getContent())
        except BaseException as e:
            errorMsg = Message.newDirectMessage(sourcePoint, Invocation.newError(message.messageID, str(e)))
            self.__stream.send_multipart(errorMsg.getContent())

    def __onResponse(self, message):
        invocation = message.getInvocation()
        correspondingID = invocation.getResponseID()
        # if isinstance(correspondingID, str): correspondingID = correspondingID.encode('UTF-8')
        self.__waitingMapLock.acquire()
        if self.__waitingMap.__contains__(correspondingID):
            (futureEntry, runnable) = self.__waitingMap.pop(correspondingID)
            if invocation.isError():
                futureEntry['error'] = invocation.getError()
            else:
                futureEntry['result'] = invocation.getResult()
            if invocation.hasWarning():
                futureEntry['warning'] = invocation.getWarning()
            runnable()
        else:
            logging.debug('ResponseID not recognized: {}'.format(message))
        self.__waitingMapLock.release()

    def send(self, msg):
        self.__stream.send_multipart(msg.getContent())
        id = msg.messageID
        (future, onFinish, resultMap) = InvokeFuture.newFuture()
        self.__waitingMapLock.acquire()
        if self.__waitingMap.__contains__(id):
            raise IFException("MessageID have been used.")
        self.__waitingMap[id] = (resultMap, onFinish)
        self.__waitingMapLock.release()
        return future

    def toMessageInvoker(self, target=None):
        return DynamicRemoteObject(None, toMessage=True, blocking=False, target=target, timeout=None)

    def asynchronousInvoker(self, target=None):
        return DynamicRemoteObject(self, toMessage=False, blocking=False, target=target, timeout=None)

    def asyncInvoker(self, target=None):
        return AsyncRemoteObject(self, target=target, timeout=30)

    def blockingInvoker(self, target=None, timeout=None):
        return DynamicRemoteObject(self, toMessage=False, blocking=True, target=target, timeout=timeout)

    @classmethod
    def start(cls):
        IOLoop.instance().start()

    def __getattr__(self, item):
        return InvokeTarget(self, item)

    def close(self):
        self.unregister()

    def bindService(self, serviceName, serviceObject, interfaces=[]):
        if self.__isService:
            raise IFException('Service already bind.')
        self.__isService = serviceName != None
        self.__serviceName = serviceName
        self.__serviceObject = serviceObject
        self.__interfaces = interfaces


class InvokeTarget:
    def __init__(self, worker, item):
        self.__worker = worker
        self.__name = item

    def __getattr__(self, item):
        item = u'{}'.format(item)
        return self.__defaultInvoker(self.__name).__getattr__(item)

    def __call__(self, *args, **kwargs):
        invoker = self.__defaultInvoker('')
        func = invoker.__getattr__(self.__name)
        return func(*args, **kwargs)

    def __defaultInvoker(self, target):
        if self.__worker.blocking:
            return self.__worker.blockingInvoker(target, self.__worker.timeout)
        else:
            return self.__worker.asynchronousInvoker(target)


class RemoteObject(object):
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return "RemoteObject[{}]".format(self.name)


class DynamicRemoteObject(RemoteObject):
    def __init__(self, worker, toMessage, blocking, target, timeout):
        super(DynamicRemoteObject, self).__init__(target)
        self.__worker = worker
        self.__target = target
        self.__toMessage = toMessage
        self.__blocking = blocking
        self.__timeout = timeout
        self.name = target

    def __getattr__(self, item):
        item = u'{}'.format(item)

        def invoke(*args, **kwargs):
            invocation = Invocation.newRequest(item, args, kwargs)
            if self.__target == '' or self.__target == None:
                message = Message.newBrokerMessage(invocation)
            else:
                message = Message.newServiceMessage(self.__target, invocation)
            if self.__toMessage:
                return message
            elif self.__blocking:
                return self.__worker.send(message).sync(self.__timeout)
            else:
                return self.__worker.send(message)

        return invoke

    def __str__(self):
        return "DynamicRemoteObject[{}]".format(self.name)


class AsyncRemoteObject(RemoteObject):
    def __init__(self, worker, target, timeout):
        super(AsyncRemoteObject, self).__init__(target)
        self.__worker = worker
        self.__target = target
        self.__timeout = timeout
        self.name = target

    def __getattr__(self, item):
        item = u'{}'.format(item)

        async def invoke(*args, **kwargs):
            invocation = Invocation.newRequest(item, args, kwargs)
            if self.__target == '' or self.__target == None:
                message = Message.newBrokerMessage(invocation)
            else:
                message = Message.newServiceMessage(self.__target, invocation)
            invokeFuture = self.__worker.send(message)
            queue = asyncio.Queue()
            aw = queue.get()

            def onComplete():
                queue.put_nowait(invokeFuture)

            invokeFuture.onComplete(onComplete)
            invokeFuture = await aw
            if invokeFuture.isSuccess():
                return invokeFuture.result()
            else:
                raise invokeFuture.exception()

        return invoke

    def __str__(self):
        return "AsyncRemoteObject[{}]".format(self.name)


class InvokeFuture:
    @classmethod
    def newFuture(cls):
        future = InvokeFuture()
        return (future, future.__onFinish, future.__resultMap)

    def __init__(self):
        self.__done = False
        self.__result = None
        self.__exception = None
        self.__warning = None
        self.__onComplete = None
        self.__metux = threading.Lock()
        self.__resultMap = {}
        self.__awaitSemaphore = threading.Semaphore(0)

    def isDone(self):
        return self.__done

    def isSuccess(self):
        return self.__exception is None

    def result(self):
        return self.__result

    def exception(self):
        return self.__exception

    def warning(self):
        return self.__warning

    def onComplete(self, func):
        self.__metux.acquire()
        self.__onComplete = func
        if self.__done:
            self.__onComplete()
        self.__metux.release()

    def waitFor(self, timeout=None):
        # For Python 3 only.
        if self.__awaitSemaphore.acquire(True, timeout):
            self.__awaitSemaphore.release()
            return True
        else:
            return False
        # For Python 2 & 3
        # timeStep = 0.1 if timeout is None else timeout / 10
        # startTime = time.time()
        # while True:
        #     acq = self.__awaitSemaphore.acquire(False)
        #     if acq:
        #         return acq
        #     else:
        #         passedTime = time.time() - startTime
        #         if (timeout is not None) and (passedTime >= timeout):
        #             return False
        #         time.sleep(timeStep)

    def sync(self, timeout=None):
        if self.waitFor(timeout):
            if self.isSuccess():
                return self.__result
            elif isinstance(self.__exception, BaseException):
                raise self.__exception
            else:
                raise IFException('Error state in InvokeFuture.')
        else:
            raise IFException('Time out!')

    def __onFinish(self):
        self.__done = True
        if self.__resultMap.__contains__('result'):
            self.__result = self.__resultMap['result']
        if self.__resultMap.__contains__('warning'):
            self.__warning = self.__resultMap['warning']
        if self.__resultMap.__contains__('error'):
            self.__exception = IFException(self.__resultMap['error'])
        if self.__onComplete is not None:
            self.__onComplete()
        self.__awaitSemaphore.release()



__license__ = "GNU General Public License v3"
__author__ = 'Hwaipy'
__email__ = 'hwaipy@gmail.com'

import zmq
from zmq.eventloop.zmqstream import ZMQStream
from interactionfreepy.core import IFDefinition, IFException, Invocation, Message, IFLoop
from tornado.ioloop import IOLoop
import time
import logging


class IFBroker(object):
    def __init__(self, binding, manager=None):
        socket = zmq.Context().socket(zmq.ROUTER)
        socket.bind(binding)
        self.main_stream = ZMQStream(socket, IOLoop.current())
        self.main_stream.on_recv(self.__onMessage)
        if manager == None:
            self.manager = Manager(self)
        else:
            self.manager = manager
        IFLoop.tryStart()

    def close(self):
        self.main_stream.on_recv(None)
        self.main_stream.socket.setsockopt(zmq.LINGER, 0)
        self.main_stream.socket.close()
        self.main_stream.close()
        self.main_stream = None

    def __onMessage(self, msg):
        try:
            sourcePoint, msg = msg[0], msg[1:]
            protocol = msg[1]
            if protocol != IFDefinition.PROTOCOL: raise IFException('Protocol {} not supported.'.format(protocol))
            distributingMode = msg[3]
            distributingAddress = msg[4]
            if distributingMode == IFDefinition.DISTRIBUTING_MODE_BROKER:
                IOLoop.current().add_callback(self.__onMessageDistributeLocal, sourcePoint, msg)
            elif distributingMode == IFDefinition.DISTRIBUTING_MODE_DIRECT:
                self.__onMessageDistributeDirect(sourcePoint, distributingAddress, msg)
            elif distributingMode == IFDefinition.DISTRIBUTING_MODE_SERVICE:
                self.__onMessageDistributeService(sourcePoint, distributingAddress, msg)
            else:
                raise IFException('Distributing mode {} not supported.'.format(distributingMode))
        except BaseException as e:
            logging.debug(e)

    async def __onMessageDistributeLocal(self, sourcePoint, msg):
        try:
            message = None
            message = Message(msg)
            invocation = message.getInvocation()
            result = self.manager.heartbeat(sourcePoint)
            if invocation.getFunction() != 'heartbeat':
                result = await invocation.perform(self.manager, sourcePoint)
            responseMessage = Message.newFromBrokerMessage(b'', Invocation.newResponse(message.messageID, result))
            self.__sendMessage([sourcePoint] + responseMessage)
        except BaseException as e:
            if message:
                errorMsg = Message.newFromBrokerMessage(b'', Invocation.newError(message.messageID, str(e)))
                self.__sendMessage([sourcePoint] + errorMsg)

    def __onMessageDistributeService(self, sourcePoint, distributingAddress, msg):
        distributingAddress = str(distributingAddress, encoding='UTF-8')
        try:
            targetAddress = self.manager.getAddressOfService(sourcePoint, distributingAddress)
            if not targetAddress: raise IFException('Service {} not exist.'.format(distributingAddress))
            self.__sendMessage([targetAddress] + msg[:3] + [sourcePoint] + msg[5:])
        except BaseException as e:
            errorMsg = Message.newFromBrokerMessage(b'', Invocation.newError(Message(msg).messageID, str(e)))
            self.__sendMessage([sourcePoint] + errorMsg)

    def __onMessageDistributeDirect(self, sourcePoint, distributingAddress, msg):
        try:
            self.__sendMessage([distributingAddress] + msg[:3] + [sourcePoint] + msg[5:])
        except BaseException as e:
            errorMsg = Message.newFromBrokerMessage(b'', Invocation.newError(Message(msg).messageID, str(e)))
            self.__sendMessage([sourcePoint] + errorMsg)

    def __sendMessage(self, frames):
        self.main_stream.send_multipart(frames)


class Manager:
    def __init__(self, broker):
        self.broker = broker
        self.__workers = {}
        self.__services = {}
        self.__activities = {}
        IOLoop.current().call_later(2, self.__check)

    def registerAsService(self, sourcePoint, name, interfaces=[]):
        if self.__services.__contains__(name):
            raise IFException('Service name [{}] occupied.'.format(name))
        if self.__workers.__contains__(sourcePoint):
            raise IFException('The current worker has registered as [{}].'.format(name))
        self.__services[name] = [sourcePoint, interfaces, time.time()]
        self.__workers[sourcePoint] = name
        logging.info('Service [{}] registered as {}.'.format(name, interfaces))

    def unregister(self, sourcePoint):
        if self.__workers.__contains__(sourcePoint):
            serviceName = self.__workers.pop(sourcePoint)
            self.__services.pop(serviceName)
            logging.info('Service [{}] unregistered.'.format(serviceName))

    def protocol(self, sourcePoint):
        return str(IFDefinition.PROTOCOL, encoding='UTF-8')

    def heartbeat(self, sourcePoint):
        self.__activities[sourcePoint] = time.time()
        return self.__workers.__contains__(sourcePoint)

    def getAddressOfService(self, sourcePoint, serviceName):
        if self.__services.__contains__(serviceName):
            return self.__services.get(serviceName)[0]
        return None

    def listServiceNames(self, sourcePoint):
        return list(self.__services.keys())

    def listServiceMeta(self, sourcePoint):
        results = []
        currentTime = time.time()
        for s in self.__services:
            meta = self.__services[s]
            results.append([s, meta[0], meta[1], currentTime - meta[2]])
        return results

    def __check(self):
        currentTime = time.time()
        for key in self.__activities.keys():
            lastActiviteTime = self.__activities[key]
            timeDiff = currentTime - lastActiviteTime
            if timeDiff > IFDefinition.HEARTBEAT_LIVETIME:
                self.unregister(key)
        IOLoop.current().call_later(2, self.__check)


if __name__ == '__main__':
    broker = IFBroker("tcp://*:2224")
    IFLoop.join()



from interactionfreepy.core import IFLoop, IFDefinition, IFException, Invocation, Message, IFLoop
from interactionfreepy.broker import IFBroker
from interactionfreepy.worker import IFWorker

import platform, asyncio
if (platform.platform().startswith('Windows')):
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()
version = '1.0.12'

setuptools.setup(
    name='interactionfreepy',
    version=version,
    author='Hwaipy',
    author_email='hwaipy@gmail.com',
    description='A intuitive and cross-languige RCP lib for Python.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='gpl-3.0',
    url='https://github.com/hwaipy/InteractionFreePy',
    download_url='https://github.com/hwaipy/InteractionFreePy/archive/v{}.tar.gz'.format(version),
    keywords=['msgpack', 'zeromq', 'zmq', '0mq', 'rcp', 'cross-languige'],
    packages=setuptools.find_packages(),
    install_requires=[
        'msgpack',
        'tornado',
        'pyzmq',
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    python_requires='>=3.6',
)


